// Copyright 2023 Greptime Team
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

//! Scans a region according to the scan request.

use std::fmt;
use std::sync::Arc;
use std::time::{Duration, Instant};

use common_error::ext::BoxedError;
use common_recordbatch::SendableRecordBatchStream;
use common_telemetry::{debug, error, warn};
use common_time::range::TimestampRange;
use common_time::Timestamp;
use datafusion::physical_plan::DisplayFormatType;
use smallvec::SmallVec;
use store_api::region_engine::RegionScannerRef;
use store_api::storage::ScanRequest;
use table::predicate::{build_time_range_predicate, Predicate};
use tokio::sync::{mpsc, Mutex, Semaphore};
use tokio_stream::wrappers::ReceiverStream;

use crate::access_layer::AccessLayerRef;
use crate::cache::file_cache::FileCacheRef;
use crate::cache::CacheManagerRef;
use crate::error::Result;
use crate::memtable::MemtableRef;
use crate::metrics::READ_SST_COUNT;
use crate::read::compat::{self, CompatBatch};
use crate::read::projection::ProjectionMapper;
use crate::read::seq_scan::SeqScan;
use crate::read::unordered_scan::UnorderedScan;
use crate::read::{Batch, Source};
use crate::region::version::VersionRef;
use crate::sst::file::{overlaps, FileHandle, FileMeta};
use crate::sst::index::applier::builder::SstIndexApplierBuilder;
use crate::sst::index::applier::SstIndexApplierRef;
use crate::sst::parquet::file_range::FileRange;

/// A scanner scans a region and returns a [SendableRecordBatchStream].
pub(crate) enum Scanner {
    /// Sequential scan.
    Seq(SeqScan),
    /// Unordered scan.
    Unordered(UnorderedScan),
}

impl Scanner {
    /// Returns a [SendableRecordBatchStream] to retrieve scan results from all partitions.
    pub(crate) async fn scan(&self) -> Result<SendableRecordBatchStream, BoxedError> {
        match self {
            Scanner::Seq(seq_scan) => seq_scan.build_stream(),
            Scanner::Unordered(unordered_scan) => unordered_scan.build_stream().await,
        }
    }

    /// Returns a [RegionScanner] to scan the region.
    pub(crate) async fn region_scanner(self) -> Result<RegionScannerRef> {
        match self {
            Scanner::Seq(seq_scan) => Ok(Arc::new(seq_scan)),
            Scanner::Unordered(unordered_scan) => Ok(Arc::new(unordered_scan)),
        }
    }
}

#[cfg(test)]
impl Scanner {
    /// Returns number of files to scan.
    pub(crate) fn num_files(&self) -> usize {
        match self {
            Scanner::Seq(seq_scan) => seq_scan.input().num_files(),
            Scanner::Unordered(unordered_scan) => unordered_scan.input().num_files(),
        }
    }

    /// Returns number of memtables to scan.
    pub(crate) fn num_memtables(&self) -> usize {
        match self {
            Scanner::Seq(seq_scan) => seq_scan.input().num_memtables(),
            Scanner::Unordered(unordered_scan) => unordered_scan.input().num_memtables(),
        }
    }

    /// Returns SST file ids to scan.
    pub(crate) fn file_ids(&self) -> Vec<crate::sst::file::FileId> {
        match self {
            Scanner::Seq(seq_scan) => seq_scan.input().file_ids(),
            Scanner::Unordered(unordered_scan) => unordered_scan.input().file_ids(),
        }
    }
}

#[cfg_attr(doc, aquamarine::aquamarine)]
/// Helper to scans a region by [ScanRequest].
///
/// [ScanRegion] collects SSTs and memtables to scan without actually reading them. It
/// creates a [Scanner] to actually scan these targets in [Scanner::scan()].
///
/// ```mermaid
/// classDiagram
/// class ScanRegion {
///     -VersionRef version
///     -ScanRequest request
///     ~scanner() Scanner
///     ~seq_scan() SeqScan
/// }
/// class Scanner {
///     <<enumeration>>
///     SeqScan
///     UnorderedScan
///     +scan() SendableRecordBatchStream
/// }
/// class SeqScan {
///     -ScanInput input
///     +build() SendableRecordBatchStream
/// }
/// class UnorderedScan {
///     -ScanInput input
///     +build() SendableRecordBatchStream
/// }
/// class ScanInput {
///     -ProjectionMapper mapper
///     -Option~TimeRange~ time_range
///     -Option~Predicate~ predicate
///     -Vec~MemtableRef~ memtables
///     -Vec~FileHandle~ files
/// }
/// class ProjectionMapper {
///     ~output_schema() SchemaRef
///     ~convert(Batch) RecordBatch
/// }
/// ScanRegion -- Scanner
/// ScanRegion o-- ScanRequest
/// Scanner o-- SeqScan
/// Scanner o-- UnorderedScan
/// SeqScan o-- ScanInput
/// UnorderedScan o-- ScanInput
/// Scanner -- SendableRecordBatchStream
/// ScanInput o-- ProjectionMapper
/// SeqScan -- SendableRecordBatchStream
/// UnorderedScan -- SendableRecordBatchStream
/// ```
pub(crate) struct ScanRegion {
    /// Version of the region at scan.
    version: VersionRef,
    /// Access layer of the region.
    access_layer: AccessLayerRef,
    /// Scan request.
    request: ScanRequest,
    /// Cache.
    cache_manager: Option<CacheManagerRef>,
    /// Parallelism to scan.
    parallelism: ScanParallism,
    /// Whether to ignore inverted index.
    ignore_inverted_index: bool,
    /// Start time of the scan task.
    start_time: Option<Instant>,
}

impl ScanRegion {
    /// Creates a [ScanRegion].
    pub(crate) fn new(
        version: VersionRef,
        access_layer: AccessLayerRef,
        request: ScanRequest,
        cache_manager: Option<CacheManagerRef>,
    ) -> ScanRegion {
        ScanRegion {
            version,
            access_layer,
            request,
            cache_manager,
            parallelism: ScanParallism::default(),
            ignore_inverted_index: false,
            start_time: None,
        }
    }

    /// Sets parallelism.
    #[must_use]
    pub(crate) fn with_parallelism(mut self, parallelism: ScanParallism) -> Self {
        self.parallelism = parallelism;
        self
    }

    #[must_use]
    pub(crate) fn with_ignore_inverted_index(mut self, ignore: bool) -> Self {
        self.ignore_inverted_index = ignore;
        self
    }

    #[must_use]
    pub(crate) fn with_start_time(mut self, now: Instant) -> Self {
        self.start_time = Some(now);
        self
    }

    /// Returns a [Scanner] to scan the region.
    pub(crate) fn scanner(self) -> Result<Scanner> {
        if self.version.options.append_mode {
            // If table uses append mode, we use unordered scan in query.
            // We still use seq scan in compaction.
            self.unordered_scan().map(Scanner::Unordered)
        } else {
            self.seq_scan().map(Scanner::Seq)
        }
    }

    /// Scan sequentially.
    pub(crate) fn seq_scan(self) -> Result<SeqScan> {
        let input = self.scan_input(true)?;
        Ok(SeqScan::new(input))
    }

    /// Unordered scan.
    pub(crate) fn unordered_scan(self) -> Result<UnorderedScan> {
        let input = self.scan_input(true)?;
        Ok(UnorderedScan::new(input))
    }

    #[cfg(test)]
    pub(crate) fn scan_without_filter_deleted(self) -> Result<SeqScan> {
        let input = self.scan_input(false)?;
        Ok(SeqScan::new(input))
    }

    /// Creates a scan input.
    fn scan_input(mut self, filter_deleted: bool) -> Result<ScanInput> {
        let time_range = self.build_time_range_predicate();

        let ssts = &self.version.ssts;
        let mut files = Vec::new();
        for level in ssts.levels() {
            for file in level.files.values() {
                // Finds SST files in range.
                if file_in_range(file, &time_range) {
                    files.push(file.clone());
                }
            }
        }

        let memtables = self.version.memtables.list_memtables();
        // Skip empty memtables and memtables out of time range.
        let memtables: Vec<_> = memtables
            .into_iter()
            .filter(|mem| {
                if mem.is_empty() {
                    return false;
                }
                let stats = mem.stats();
                // Safety: the memtable is not empty.
                let (start, end) = stats.time_range().unwrap();

                // The time range of the memtable is inclusive.
                let memtable_range = TimestampRange::new_inclusive(Some(start), Some(end));
                memtable_range.intersects(&time_range)
            })
            .collect();

        debug!(
            "Scan region {}, request: {:?}, memtables: {}, ssts_to_read: {}, append_mode: {}",
            self.version.metadata.region_id,
            self.request,
            memtables.len(),
            files.len(),
            self.version.options.append_mode,
        );

        let index_applier = self.build_index_applier();
        let predicate = Predicate::new(self.request.filters.clone());
        // The mapper always computes projected column ids as the schema of SSTs may change.
        let mapper = match &self.request.projection {
            Some(p) => ProjectionMapper::new(&self.version.metadata, p.iter().copied())?,
            None => ProjectionMapper::all(&self.version.metadata)?,
        };

        let input = ScanInput::new(self.access_layer, mapper)
            .with_time_range(Some(time_range))
            .with_predicate(Some(predicate))
            .with_memtables(memtables)
            .with_files(files)
            .with_cache(self.cache_manager)
            .with_index_applier(index_applier)
            .with_parallelism(self.parallelism)
            .with_start_time(self.start_time)
            .with_append_mode(self.version.options.append_mode)
            .with_filter_deleted(filter_deleted);
        Ok(input)
    }

    /// Build time range predicate from filters.
    fn build_time_range_predicate(&mut self) -> TimestampRange {
        let time_index = self.version.metadata.time_index_column();
        let unit = time_index
            .column_schema
            .data_type
            .as_timestamp()
            .expect("Time index must have timestamp-compatible type")
            .unit();
        build_time_range_predicate(
            &time_index.column_schema.name,
            unit,
            &mut self.request.filters,
        )
    }

    /// Use the latest schema to build the index applier.
    fn build_index_applier(&self) -> Option<SstIndexApplierRef> {
        if self.ignore_inverted_index {
            return None;
        }

        let file_cache = || -> Option<FileCacheRef> {
            let cache_manager = self.cache_manager.as_ref()?;
            let write_cache = cache_manager.write_cache()?;
            let file_cache = write_cache.file_cache();
            Some(file_cache)
        }();

        SstIndexApplierBuilder::new(
            self.access_layer.region_dir().to_string(),
            self.access_layer.object_store().clone(),
            file_cache,
            self.version.metadata.as_ref(),
            self.version
                .options
                .index_options
                .inverted_index
                .ignore_column_ids
                .iter()
                .copied()
                .collect(),
        )
        .build(&self.request.filters)
        .inspect_err(|err| warn!(err; "Failed to build index applier"))
        .ok()
        .flatten()
        .map(Arc::new)
    }
}

/// Config for parallel scan.
#[derive(Debug, Clone, Copy, Default)]
pub(crate) struct ScanParallism {
    /// Number of tasks expect to spawn to read data.
    pub(crate) parallelism: usize,
    /// Channel size to send batches. Only takes effect when the parallelism > 1.
    pub(crate) channel_size: usize,
}

/// Returns true if the time range of a SST `file` matches the `predicate`.
fn file_in_range(file: &FileHandle, predicate: &TimestampRange) -> bool {
    if predicate == &TimestampRange::min_to_max() {
        return true;
    }
    // end timestamp of a SST is inclusive.
    let (start, end) = file.time_range();
    let file_ts_range = TimestampRange::new_inclusive(Some(start), Some(end));
    file_ts_range.intersects(predicate)
}

/// Common input for different scanners.
pub(crate) struct ScanInput {
    /// Region SST access layer.
    access_layer: AccessLayerRef,
    /// Maps projected Batches to RecordBatches.
    pub(crate) mapper: Arc<ProjectionMapper>,
    /// Time range filter for time index.
    time_range: Option<TimestampRange>,
    /// Predicate to push down.
    pub(crate) predicate: Option<Predicate>,
    /// Memtables to scan.
    pub(crate) memtables: Vec<MemtableRef>,
    /// Handles to SST files to scan.
    pub(crate) files: Vec<FileHandle>,
    /// Cache.
    pub(crate) cache_manager: Option<CacheManagerRef>,
    /// Ignores file not found error.
    ignore_file_not_found: bool,
    /// Parallelism to scan data.
    pub(crate) parallelism: ScanParallism,
    /// Index applier.
    index_applier: Option<SstIndexApplierRef>,
    /// Start time of the query.
    pub(crate) query_start: Option<Instant>,
    /// The region is using append mode.
    pub(crate) append_mode: bool,
    /// Whether to remove deletion markers.
    pub(crate) filter_deleted: bool,
}

impl ScanInput {
    /// Creates a new [ScanInput].
    #[must_use]
    pub(crate) fn new(access_layer: AccessLayerRef, mapper: ProjectionMapper) -> ScanInput {
        ScanInput {
            access_layer,
            mapper: Arc::new(mapper),
            time_range: None,
            predicate: None,
            memtables: Vec::new(),
            files: Vec::new(),
            cache_manager: None,
            ignore_file_not_found: false,
            parallelism: ScanParallism::default(),
            index_applier: None,
            query_start: None,
            append_mode: false,
            filter_deleted: true,
        }
    }

    /// Sets time range filter for time index.
    #[must_use]
    pub(crate) fn with_time_range(mut self, time_range: Option<TimestampRange>) -> Self {
        self.time_range = time_range;
        self
    }

    /// Sets predicate to push down.
    #[must_use]
    pub(crate) fn with_predicate(mut self, predicate: Option<Predicate>) -> Self {
        self.predicate = predicate;
        self
    }

    /// Sets memtables to read.
    #[must_use]
    pub(crate) fn with_memtables(mut self, memtables: Vec<MemtableRef>) -> Self {
        self.memtables = memtables;
        self
    }

    /// Sets files to read.
    #[must_use]
    pub(crate) fn with_files(mut self, files: Vec<FileHandle>) -> Self {
        self.files = files;
        self
    }

    /// Sets cache for this query.
    #[must_use]
    pub(crate) fn with_cache(mut self, cache: Option<CacheManagerRef>) -> Self {
        self.cache_manager = cache;
        self
    }

    /// Ignores file not found error.
    #[must_use]
    pub(crate) fn with_ignore_file_not_found(mut self, ignore: bool) -> Self {
        self.ignore_file_not_found = ignore;
        self
    }

    /// Sets scan parallelism.
    #[must_use]
    pub(crate) fn with_parallelism(mut self, parallelism: ScanParallism) -> Self {
        self.parallelism = parallelism;
        self
    }

    /// Sets index applier.
    #[must_use]
    pub(crate) fn with_index_applier(mut self, index_applier: Option<SstIndexApplierRef>) -> Self {
        self.index_applier = index_applier;
        self
    }

    /// Sets start time of the query.
    #[must_use]
    pub(crate) fn with_start_time(mut self, now: Option<Instant>) -> Self {
        self.query_start = now;
        self
    }

    #[must_use]
    pub(crate) fn with_append_mode(mut self, is_append_mode: bool) -> Self {
        self.append_mode = is_append_mode;
        self
    }

    /// Sets whether to remove deletion markers during scan.
    #[must_use]
    pub(crate) fn with_filter_deleted(mut self, filter_deleted: bool) -> Self {
        self.filter_deleted = filter_deleted;
        self
    }

    /// Scans sources in parallel.
    ///
    /// # Panics if the input doesn't allow parallel scan.
    pub(crate) fn create_parallel_sources(
        &self,
        sources: Vec<Source>,
        semaphore: Arc<Semaphore>,
    ) -> Result<Vec<Source>> {
        debug_assert!(self.parallelism.parallelism > 1);
        // Spawn a task for each source.
        let sources = sources
            .into_iter()
            .map(|source| {
                let (sender, receiver) = mpsc::channel(self.parallelism.channel_size);
                self.spawn_scan_task(source, semaphore.clone(), sender);
                let stream = Box::pin(ReceiverStream::new(receiver));
                Source::Stream(stream)
            })
            .collect();
        Ok(sources)
    }

    /// Prunes file ranges to scan and adds them to the `collector`.
    pub(crate) async fn prune_file_ranges(
        &self,
        collector: &mut impl FileRangeCollector,
    ) -> Result<()> {
        for file in &self.files {
            let res = self
                .access_layer
                .read_sst(file.clone())
                .predicate(self.predicate.clone())
                .time_range(self.time_range)
                .projection(Some(self.mapper.column_ids().to_vec()))
                .cache(self.cache_manager.clone())
                .index_applier(self.index_applier.clone())
                .expected_metadata(Some(self.mapper.metadata().clone()))
                .build_reader_input()
                .await;
            let (mut file_range_ctx, row_groups) = match res {
                Ok(x) => x,
                Err(e) => {
                    if e.is_object_not_found() && self.ignore_file_not_found {
                        error!(e; "File to scan does not exist, region_id: {}, file: {}", file.region_id(), file.file_id());
                        continue;
                    } else {
                        return Err(e);
                    }
                }
            };
            if !compat::has_same_columns(
                self.mapper.metadata(),
                file_range_ctx.read_format().metadata(),
            ) {
                // They have different schema. We need to adapt the batch first so the
                // mapper can convert it.
                let compat = CompatBatch::new(
                    &self.mapper,
                    file_range_ctx.read_format().metadata().clone(),
                )?;
                file_range_ctx.set_compat_batch(Some(compat));
            }
            // Build ranges from row groups.
            let file_range_ctx = Arc::new(file_range_ctx);
            let file_ranges = row_groups
                .into_iter()
                .map(|(row_group_idx, row_selection)| {
                    FileRange::new(file_range_ctx.clone(), row_group_idx, row_selection)
                });
            collector.append_file_ranges(file.meta_ref(), file_ranges);
        }

        READ_SST_COUNT.observe(self.files.len() as f64);

        Ok(())
    }

    /// Scans the input source in another task and sends batches to the sender.
    pub(crate) fn spawn_scan_task(
        &self,
        mut input: Source,
        semaphore: Arc<Semaphore>,
        sender: mpsc::Sender<Result<Batch>>,
    ) {
        common_runtime::spawn_read(async move {
            loop {
                // We release the permit before sending result to avoid the task waiting on
                // the channel with the permit held.
                let maybe_batch = {
                    // Safety: We never close the semaphore.
                    let _permit = semaphore.acquire().await.unwrap();
                    input.next_batch().await
                };
                match maybe_batch {
                    Ok(Some(batch)) => {
                        let _ = sender.send(Ok(batch)).await;
                    }
                    Ok(None) => break,
                    Err(e) => {
                        let _ = sender.send(Err(e)).await;
                        break;
                    }
                }
            }
        });
    }
}

#[cfg(test)]
impl ScanInput {
    /// Returns number of memtables to scan.
    pub(crate) fn num_memtables(&self) -> usize {
        self.memtables.len()
    }

    /// Returns number of SST files to scan.
    pub(crate) fn num_files(&self) -> usize {
        self.files.len()
    }

    /// Returns SST file ids to scan.
    pub(crate) fn file_ids(&self) -> Vec<crate::sst::file::FileId> {
        self.files.iter().map(|file| file.file_id()).collect()
    }
}

/// Groups of file ranges. Each group in the list contains multiple file
/// ranges to scan. File ranges in the same group may come from different files.
pub(crate) type FileRangesGroup = SmallVec<[Vec<FileRange>; 4]>;

/// A partition of a scanner to read.
/// It contains memtables and file ranges to scan.
#[derive(Default)]
pub(crate) struct ScanPart {
    /// Memtables to scan.
    /// We scan the whole memtable now. We might scan a range of the memtable in the future.
    pub(crate) memtables: Vec<MemtableRef>,
    /// File ranges to scan.
    pub(crate) file_ranges: FileRangesGroup,
    /// Optional time range of the part (inclusive).
    pub(crate) time_range: Option<(Timestamp, Timestamp)>,
}

impl fmt::Debug for ScanPart {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "ScanPart({} memtables, {} file ranges",
            self.memtables.len(),
            self.file_ranges
                .iter()
                .map(|ranges| ranges.len())
                .sum::<usize>(),
        )?;
        if let Some(time_range) = &self.time_range {
            write!(f, ", time range: {:?})", time_range)
        } else {
            write!(f, ")")
        }
    }
}

impl ScanPart {
    /// Returns true if the time range given `part` overlaps with this part.
    pub(crate) fn overlaps(&self, part: &ScanPart) -> bool {
        let (Some(current_range), Some(part_range)) = (self.time_range, part.time_range) else {
            return true;
        };

        overlaps(&current_range, &part_range)
    }

    /// Merges given `part` to this part.
    pub(crate) fn merge(&mut self, mut part: ScanPart) {
        self.memtables.append(&mut part.memtables);
        self.file_ranges.append(&mut part.file_ranges);
        let Some(part_range) = part.time_range else {
            return;
        };
        let Some(current_range) = self.time_range else {
            self.time_range = part.time_range;
            return;
        };
        let start = current_range.0.min(part_range.0);
        let end = current_range.1.max(part_range.1);
        self.time_range = Some((start, end));
    }

    /// Returns true if the we can split the part into multiple parts
    /// and preserving order.
    pub(crate) fn can_split_preserve_order(&self) -> bool {
        self.memtables.is_empty() && self.file_ranges.len() == 1 && self.file_ranges[0].len() > 1
    }
}

/// A trait to collect file ranges to scan.
pub(crate) trait FileRangeCollector {
    /// Appends file ranges from the **same file** to the collector.
    fn append_file_ranges(
        &mut self,
        file_meta: &FileMeta,
        file_ranges: impl Iterator<Item = FileRange>,
    );
}

/// Optional list of [ScanPart]s.
#[derive(Default)]
pub(crate) struct ScanPartList(pub(crate) Option<Vec<ScanPart>>);

impl fmt::Debug for ScanPartList {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match &self.0 {
            Some(parts) => write!(f, "{:?}", parts),
            None => write!(f, "[]"),
        }
    }
}

impl ScanPartList {
    /// Returns true if the list is None.
    pub(crate) fn is_none(&self) -> bool {
        self.0.is_none()
    }

    /// Sets parts to the list.
    pub(crate) fn set_parts(&mut self, parts: Vec<ScanPart>) {
        self.0 = Some(parts);
    }

    /// Gets the part by index, returns None if the index is out of bound.
    /// # Panics
    /// Panics if parts are not initialized.
    pub(crate) fn get_part(&mut self, index: usize) -> Option<&ScanPart> {
        let parts = self.0.as_ref().unwrap();
        parts.get(index)
    }

    /// Returns the number of parts.
    pub(crate) fn len(&self) -> usize {
        self.0.as_ref().map_or(0, |parts| parts.len())
    }

    /// Returns the number of memtables.
    pub(crate) fn num_memtables(&self) -> usize {
        self.0.as_ref().map_or(0, |parts| {
            parts.iter().map(|part| part.memtables.len()).sum()
        })
    }

    /// Returns the number of file ranges.
    pub(crate) fn num_file_ranges(&self) -> usize {
        self.0.as_ref().map_or(0, |parts| {
            parts.iter().map(|part| part.file_ranges.len()).sum()
        })
    }
}

/// Context shared by different streams from a scanner.
/// It contains the input and distributes input to multiple parts
/// to scan.
pub(crate) struct StreamContext {
    /// Input memtables and files.
    pub(crate) input: ScanInput,
    /// Parts to scan.
    /// The scanner builds parts to scan from the input lazily.
    /// The mutex is used to ensure the parts are only built once.
    pub(crate) parts: Mutex<ScanPartList>,

    // Metrics:
    /// The start time of the query.
    pub(crate) query_start: Instant,
    /// Time elapsed before creating the scanner.
    pub(crate) prepare_scan_cost: Duration,
}

impl StreamContext {
    /// Creates a new [StreamContext].
    pub(crate) fn new(input: ScanInput) -> Self {
        let query_start = input.query_start.unwrap_or_else(Instant::now);
        let prepare_scan_cost = query_start.elapsed();

        Self {
            input,
            parts: Mutex::new(ScanPartList::default()),
            query_start,
            prepare_scan_cost,
        }
    }

    /// Format parts for explain.
    pub(crate) fn format_parts(&self, t: DisplayFormatType, f: &mut fmt::Formatter) -> fmt::Result {
        match self.parts.try_lock() {
            Ok(inner) => match t {
                DisplayFormatType::Default => write!(
                    f,
                    "partition_count={} ({} memtables, {} file ranges)",
                    inner.len(),
                    inner.num_memtables(),
                    inner.num_file_ranges()
                ),
                DisplayFormatType::Verbose => write!(f, "{:?}", &*inner),
            },
            Err(_) => write!(f, "<locked>"),
        }
    }
}
